/* Written by Steven Kilthau */

#ifndef _VECTOR_H_
#define _VECTOR_H_

#include <iostream>

#include "vuSimpleTypes.h"

class vuVector; //forward declaration.
typedef vuVector* vuVectorPointer;
typedef vuVectorPointer* vuVectorPointerArray;

class vuMatrix;

//!Taken from vuVolume2. A class for representing 3d vectors.
/*!
  The vuVector class is a 3D homogeneous vector
  that works in conjunction with the vuMatrix
  class (a 4x4 matrix).  The elements of the
  vuVector are all single precision floating-
  point numbers.
*/

class vuVector
{
friend class vuMatrix;
public:
    //!Default constructor making a zero vector.
    vuVector();
    //!Copy constructor.
    vuVector(const vuVector& v);
    //!Constructor initializing to a 3d vector in homogenous space.
    vuVector(double v1, double v2, double v3);
    //!Constructor initializing to a homogenous 3d vector.
    vuVector(double v1, double v2, double v3, double v4);
    //!Constructor initializing the x,y,z coordinates to v.
    vuVector(double v);
    //!Constructor initializing the vector from an array of doubles.
    /*!\param v An array of 3 doubleing point values corresponding to x,y,z.
    */
    vuVector(const double* v);
    //!The destructor.
    ~vuVector();

    //!The assignment operator.
    vuVector& operator=(const vuVector& v);
    //!Assigns the value v to the x,y,z elements of the vector.
    vuVector& operator=(double v);
    //!Assigns a 3d double array to the vector.
    vuVector& operator=(const double* v);

    //!Computes the 2-norm on the vector (lenght operator).
    double norm(void) const;
    //!Computes squared 2-norm (scalar product with itself).
    double norm2(void) const;

    //!Normalizes the vector to a length of 1.
    /*!This makes a unit vector which points in the same direction as the instance.
    */
    vuVector& makeUnit(void);
    //!Normalizes the vector in homogenous space.
    /*!Divides the whole vector by the 4th coordinate to make it 1.
    */
    vuVector& normalize(void);

    //!Returns the entry-wise inverse of the vector.
    /*!Entry wise meaning that each of the x, y, z coordinates is treated separately,
       leaving the fourth w coordinate untouched.
    */
    vuVector inv(void) const;
    //!Inverts the vector entry-wise.
    vuVector& invEq(void);
    //!Returns the entry-wise product of the vector with another vector.
    vuVector mul(const vuVector& rhs) const;
    //!Multiplies the vector, entry-wise, by another vector.
    vuVector& mulEq(const vuVector& rhs);
    //!Returns the entry-wise division of the vector by another vector.
    vuVector div(const vuVector& rhs) const;
    //!Divides the vector, entry-wise, by another vector.
    vuVector& divEq(const vuVector& rhs);

    //!The access operator.
    double& operator[](dword index);
    //!The const access operator.
    const double& operator[](dword index) const;

    //!Returns the data pointer.
    /*!The data is stored as a four dimensional double array.
    	The getData() function can both retrieve and modify; it
     	is similar to the [] operator. To warrant the return of
      a const pointer (for use in other const member functions)
      explicitly cast the pointer returned to (const).
    */
    double* getData(void);
    //! const version of getData()
    double const* getData(void) const;
    
    //!Returns the dot product of two vectors.
    friend double dot(const vuVector& v1,const vuVector& v2);
    /** Returns the (LH) cross product of two vectors.
	Left-handed means use left hand and the correspondence is
	v1-thumb, v2-pointing finger, result-middle finger
     */
    friend vuVector cross(const vuVector& v1,const vuVector& v2);
    //!Returns the dot product of the vector with another vector.
    double dot(const vuVector& v) const;
    /** Returns the (LH) cross product of the vector with another vector.
	Left-handed means use left hand and the correspondence is
	*this-thumb, v-pointing finger, result-middle finger
     */
    vuVector cross(const vuVector& v) const;

    //!Returns the sum of the vector with another vector.
    vuVector operator+(const vuVector& v) const;
    //!Returns the difference of the vector from another vector.
    vuVector operator-(const vuVector& v) const;
    //!Returns the product of the vector and a matrix.
    vuVector operator*(const vuMatrix& m) const;
    //!Returns the product of the vector with another vector.
    vuMatrix operator*(const vuVector& v) const;
    //!Returns the product of the vector with a scalar.
    vuVector operator*(double s) const;
    //!Returns the fraction of the vector with a scalar
    vuVector vuVector::operator/(double s) const;
    //!Returns the product of a scalar with a vector.
    friend const vuVector operator*(double s, const vuVector& v);
    //!Adds a vector to this vector.
    vuVector& operator+=(const vuVector& v);
    //!Subtracts a vector from this vector.
    vuVector& operator-=(const vuVector& v);
    //!Multiplies the vector by a matrix.
    vuVector& operator*=(const vuMatrix& m);
    //!Multiplies the vector by a scalar.
    vuVector& operator*=(double s);

    //!An equality operator, comparing only the x,y,z coordinates.
    bool operator==(const vuVector& v) const;
    //!An inequality operator, comparing only the x,y,z coordinates.
    bool operator!=(const vuVector& v) const;

    //!Returns index of component with highest absolute
    int getDominantAxis() const;

    //! writes the foure elements to a stream; separated by whitespaces
    friend std::ostream& operator<<(std::ostream& out,const vuVector& v);
    //! reads a vector from a stream
    friend std::istream& operator>>(std::istream& in, vuVector& v);

private:
    //!The coordinates of the vector stored as a doubleing point array.
    double val[4];
};

#endif
