/****************************************************************************

Author: Tai Meng

compiled and tested on amber.csil.sfu.ca

Usage: > ./denoise [dir] [filename] [noiseType] [gridType] [filterRadius]
         - [dir] 
           - the folder where the noisy dataset is found
           - the folder where the denoised dataset is written to
         - [filename] 
           - the name of the noisy dataset file
           - ".dat" is appended automatically.
           - [filename].dat must be a double-precision dataset
         - [noiseType]
           - 0: Salt & pepper noise
           - 1: Gaussian noise
         - [gridType]
           - 0: CC
           - 1: BCC
         - [filterRadius]
           - radius of the continuous spherical filter 
         - Output the denoised dataset in:
           - [dir]/[filename]_denoise.dat

****************************************************************************/

//============================================================
//Includes
//============================================================

#include <cmath>
#include <assert.h>
#include <cstdlib>

#include <iostream>
#include <string>
using namespace std;

#include <vuML.h>
#include <vuVector.h>
#include <Dataset.h>
#include <Utilities.h>
#include <MedianFilter.h>

//============================================================
//Global objects
//============================================================
Dataset dataset_noisy;
Dataset dataset_denoised;

MedianSphericalFilterCC median_filter_cc;
MedianSphericalFilterBCC median_filter_bcc;
GaussianSphericalFilterCC gaussian_filter_cc;
GaussianSphericalFilterBCC gaussian_filter_bcc;
SphericalFilter* active_filter = NULL;

string dir = "../common/data/";
string filename;

enum NoiseType {SALT, GAUSSIAN};
NoiseType noiseType = SALT;

GridType gridType = CC;

double filterRadius = 0.1;

//============================================================
//Functions
//============================================================

void denoise() {
  if ( noiseType == SALT ) {
    if ( gridType == CC ) active_filter = &median_filter_cc;
    else active_filter = &median_filter_bcc;
  }
  if ( noiseType == GAUSSIAN ) {
    if ( gridType == CC ) active_filter = &gaussian_filter_cc;
    else active_filter = &gaussian_filter_bcc;
  }

  assert(filterRadius >= 0.0);
  active_filter->setRadius(filterRadius);

  string fullname = dir+filename+string(".dat");
  dataset_noisy.read(fullname);
  assert(dataset_noisy.getGridType() == gridType);
  active_filter->filter(dataset_noisy, dataset_denoised);

  dataset_denoised.write(dir, filename+string("_denoised"), ".vud");
}

int main(int argc, char** argv)
{
  if ( argc >= 6 ) {
    dir = string(argv[1]);
    filename = string(argv[2]);
    noiseType = NoiseType(atoi(argv[3]));
    gridType = GridType(atoi(argv[4]));
    filterRadius = atof(argv[5]);
  }
  else {
    cout << "Some parameters are missing or incorrect. Please try again. " << endl;
    return -1;
  }

  denoise();

  return 0;
}
