/****************************************************************************

Author: Tai Meng

compiled and tested on amber.csil.sfu.ca

Usage: > ./sampler [CC_dim] [stdDev] [probability] [outDir]
         - [CC_dim] 
           - the number of samples along any major axis on the Cartesian grid
           - the BCC dimensions are calculated so that BCC has 70% of samples of CC
         - [stdDev] 
           - the standard deviation of the 0 meaned white Gaussian noise
         - [probability] 
           - the probability that a sample becomes affected by salt & pepper noise
         - [outDir] 
           - the folder in which samplings (datasets) are stored
         - Output in the [outDir] folder: 10 datasets in total
           - one set in bytes (*.vud), and one set in doubles (*.dat) file format:
             - pair of noise-free datasets
             - pair of datasets with gaussian noise for CC
             - pair of datasets with gaussian noise for BCC
             - pair of datasets with salt & pepper nosie for CC
             - pair of datasets with salt & pepper nosie for BCC

****************************************************************************/

//============================================================
//Includes
//============================================================

#include <cmath>
#include <assert.h>
#include <cstdlib>
#include <iostream>
#include <string>
using namespace std;

#include <vuML.h>
#include <vuVector.h>
#include <Dataset.h>
#include <Utilities.h>

//============================================================
//Global objects
//============================================================
Dataset CC100(CC);
Dataset BCC70(BCC);
Dataset CC100_saltPepperNoise(CC);
Dataset BCC70_saltPepperNoise(BCC);
Dataset CC100_gaussian(CC);
Dataset BCC70_gaussian(BCC);
MarschnerLobb ml;

int CC100_dim;
double stdDev;
double probability;
string out_dir;

//============================================================
//Functions
//============================================================

//this function works for all integer coords, + or -
vuVector convertBCCIndex(int x, int y, int z) {
  Vector3i index_bcc = Dataset::index_cc_to_bcc(x,y,z);
  vuVector iv(index_bcc[0], index_bcc[1], index_bcc[2]);
  return iv;
}

void printNoiseLevels() {
  double percentage, mean, stdDev;

  cout << "===================== CC salt ==============================" << endl;
  Dataset::saltPepperError(CC100, CC100_saltPepperNoise, percentage);
  cout << "percentage = " << percentage << endl;

  cout << "===================== BCC salt =============================" << endl;
  Dataset::saltPepperError(BCC70, BCC70_saltPepperNoise, percentage);
  cout << "percentage = " << percentage << endl;

  cout << "===================== CC gauss =============================" << endl;
  Dataset::whiteNoiseError(CC100, CC100_gaussian, mean, stdDev);
  cout << "mean = " << mean << endl;
  cout << "stdDev = " << stdDev << endl;

  cout << "======================BCC gauss=============================" << endl;
  Dataset::whiteNoiseError(BCC70, BCC70_gaussian, mean , stdDev);
  cout << "mean = " << mean << endl;
  cout << "stdDev = " << stdDev << endl;
}

void writeVolumes(){
  CC100.write(out_dir, "01_ML_CC100", ".vud");
  BCC70.write(out_dir, "02_ML_BCC70", ".vud");
  CC100_saltPepperNoise.write(out_dir, "03_ML_CC100_salt_pepper", ".vud");
  BCC70_saltPepperNoise.write(out_dir, "04_ML_BCC70_salt_pepper", ".vud");
  CC100_gaussian.write(out_dir, "05_ML_CC100_gaussian", ".vud");
  BCC70_gaussian.write(out_dir, "06_ML_BCC70_gaussian", ".vud");
}

void fillBCC70() {
  int xmax = BCC70.getXDim();
  int ymax = BCC70.getYDim();
  int zmax = BCC70.getZDim();

  int x, y, z;
  for ( z = 0; z < zmax; z++ ) {
    for ( y = 0; y < ymax; y++ ) {
      for (x = 0; x < xmax; x++ ) {
	int i = BCC70.lindex_cc(x, y, z);
	vuVector p = convertBCCIndex(x, y, z);
	p.mulEq(BCC70.getScale()); //scale this point
	BCC70[i] = ml.getValue(p);
	BCC70_saltPepperNoise[i] = ml.getValueSaltPepperNoise(p, probability);
	BCC70_gaussian[i] = ml.getValueGaussianNoise(p, stdDev);
      }
    }
  }
}

void fillCC100() {
  int xmax = CC100.getXDim();
  int ymax = CC100.getYDim();
  int zmax = CC100.getZDim();

  int x, y, z;
  for ( z = 0; z < zmax; z++ ) {
    for ( y = 0; y < ymax; y++ ) {
      for (x = 0; x < xmax; x++ ) {
	int i = CC100.lindex_cc(x, y, z);
	vuVector p = vuVector(double(x), double(y), double(z));
	CC100[i] = ml.getValue(p);
	CC100_saltPepperNoise[i] = ml.getValueSaltPepperNoise(p, probability);
	CC100_gaussian[i] = ml.getValueGaussianNoise(p, stdDev);
      }
    }
  }
}

void fillVolumes() {
  fillCC100();
  fillBCC70();
}

void initVolumes() {
  CC100.setDim(CC100_dim, CC100_dim, CC100_dim);
  CC100_saltPepperNoise.setDim(CC100_dim, CC100_dim, CC100_dim);
  CC100_gaussian.setDim(CC100_dim, CC100_dim, CC100_dim);
  
  //BCC70_dim^3*2 = CC100_dataSize / sqrt(2)
  int CC100_dataSize = CC100.getNumSamples();
  int BCC70_dim = myRound(pow(CC100_dataSize/sqrt(2.)/2., 1./3.));
  BCC70.setDim(BCC70_dim, BCC70_dim, BCC70_dim*2);    
  BCC70_saltPepperNoise.setDim(BCC70_dim, BCC70_dim, BCC70_dim*2);    
  BCC70_gaussian.setDim(BCC70_dim, BCC70_dim, BCC70_dim*2);    
  
  //scale BCC's bounding box to match cartesian's
  //CC bounding box: (CC100_dim - 1)^3
  //BCC bounding box: (BCC70_dim*2 - 1)^3
  double bcc_scale = (CC100_dim - 1.0) / (BCC70_dim*2.0 - 1.0);
  BCC70.setScaling( bcc_scale );  
  BCC70_saltPepperNoise.setScaling( bcc_scale );  
  BCC70_gaussian.setScaling( bcc_scale );  
  //CC scaling defaulted to 1

  printf("CC dimensions: %i ^ 3\n" 
	 "BCC dimensions: %i, %i, %i\n", CC100_dim, BCC70_dim, BCC70_dim, BCC70_dim*2);
  cout << "CC scaling: " << CC100.getScale() << endl;
  cout << "BCC scaling: " << BCC70.getScale() << endl;

  //initialize ML volume to match bounding box of CC100
  ml.initialize(CC100_dim, CC100_dim, CC100_dim);
}

int main(int argc, char** argv)
{
  if ( argc >= 5 ) {
    CC100_dim = atoi(argv[1]);
    stdDev = atof(argv[2]);
    probability = atof(argv[3]);
    out_dir = string(argv[4]);
  }    
  else {
    cout << "Some parameters are missing or incorrect. Please try again. " << endl;
    return -1;
  }
    
  initVolumes();
  fillVolumes();
  writeVolumes();
  printNoiseLevels();
    
  return 0;
}
