/****************************************************************************

Author: Tai Meng

compiled and tested on amber.csil.sfu.ca

Usage: > ./plotData [dir] [data_orig.dat] [data_noisy.dat] [plotData.dat] [noiseType] [gridType] [isWithBoundary]
         - [dir] 
           - the folder where the datasets are found
           - the folder where the plot data is written to
         - [data_orig.dat]
           - name of noise-free dataset
         - [data_noisy.dat]
           - name of corresponding noisy dataset
         - [plotData.dat]
           - name of plot data file
         - [noiseType]
           - 0: Salt & pepper noise
           - 1: Gaussian noise
         - [gridType]
           - 0: CC
           - 1: BCC
         - [isWithBoundary]
           - 0: consider only interior samples
           - 1: consider all samples
       - output in:
         - [dir]/[plotData.dat]
         - data points in ASCII for plotting in Matlab
         - data file format
           - line 1: numCols numRows (i.e. 4 10)
           - subsequent numRows lines: 1 line (or row) per entry
           - each entry has format
             - filterRadius neighborhoodSize mean stdDev
             - i.e. 1.000 7 0.01 2.0

****************************************************************************/

//============================================================
//Includes
//============================================================

#include <cmath>
#include <assert.h>
#include <cstdlib>

#include <iostream>
#include <string>
using namespace std;

#include <vuML.h>
#include <vuVector.h>
#include <Dataset.h>
#include <Utilities.h>
#include <MedianFilter.h>
#include <VVector.h>

//============================================================
//Global objects
//============================================================
Dataset dataset_orig;
Dataset dataset_noisy;
Dataset dataset_denoised;

MedianSphericalFilterCC median_filter_cc;
MedianSphericalFilterBCC median_filter_bcc;
GaussianSphericalFilterCC gaussian_filter_cc;
GaussianSphericalFilterBCC gaussian_filter_bcc;
SphericalFilter* active_filter = NULL;

string dir;
string filename_orig;
string filename_noisy;
string filename_out;

enum NoiseType {SALT, GAUSSIAN};
NoiseType noiseType;
GridType gridType;
bool isWithBoundary;

//============================================================
//Functions
//============================================================

void chooseFilter() {
 if ( noiseType == SALT ) {
    if ( gridType == CC ) active_filter = &median_filter_cc;
    else active_filter = &median_filter_bcc;
  }
  if ( noiseType == GAUSSIAN ) {
    if ( gridType == CC ) active_filter = &gaussian_filter_cc;
    else active_filter = &gaussian_filter_bcc;
  }
}

double getScale() {
  if ( gridType == CC ) return 1.0;

  //scale for CC 64^3 vs BCC 45x45x90
  //theoretically this number is sqrt(2.0)
  //but in practice 1 unit of distance in CC needs to be 
  //equal to 1 unit of distance in BCC, so slightly
  //different from 2.0
  return 89.0/63.0;
}

void writeEntry(FILE* fout) {
  assert(fout);
  
  double radius = active_filter->getRadius();
  double scale = getScale();
  int neighborhoodSize = active_filter->getNeighborhoodSize();
  double mean, stdDev;

  if ( isWithBoundary ) {
    Dataset::denoiseErrorWithBoundaries(dataset_orig,
					dataset_denoised,
					mean, stdDev);
  }
  else { //exclude boundary
    double filterRadius = active_filter->getRadius();
    Dataset::denoiseErrorWithoutBoundaries(dataset_orig,
					   dataset_denoised,
					   mean, stdDev,
					   filterRadius);
  }
  fprintf(fout, "%lf\t%i\t%lf\t%lf\n",
	  radius/scale, neighborhoodSize,
	  mean, stdDev);
}

void readDatasets() {
  dataset_orig.read(dir+filename_orig);
  assert(gridType == dataset_orig.getGridType());

  dataset_noisy.read(dir+filename_noisy);
  assert(gridType == dataset_noisy.getGridType());
}

void writePlotData(FILE* fout) {
  //first write header: numCols numRows
  const int TOTAL_NUM_ENTRIES = 10; //this many neighborhoods to plot
  //[radius, #neighbors, mean, variance], 4 components per entry
  fprintf(fout, "%i\t%i\n", 4, TOTAL_NUM_ENTRIES); 

  int num_entries_so_far = 0;
  int radiusSquared = 0;
  int prevNeighborhoodSize = 0;
  int currentNeighborhoodSize;
  while ( num_entries_so_far < TOTAL_NUM_ENTRIES ) {
    double radius = sqrt(double(radiusSquared));
    active_filter->setRadius(radius);

    currentNeighborhoodSize = active_filter->getNeighborhoodSize();
    
    //avoid duplicate entries
    if ( currentNeighborhoodSize > prevNeighborhoodSize ) {
      assert(EQUALS(radius, active_filter->getRadius()));      
      double scale = getScale();
      assert(radius/scale <= 16.0);

      active_filter->filter(dataset_noisy, dataset_denoised);
      writeEntry(fout);
      num_entries_so_far++;

      prevNeighborhoodSize = currentNeighborhoodSize;
    }
    
    //TM: should output 1 intermediate data to see correctness
    //dataset_denoised.write(dir, filename_out, ".vud");

    radiusSquared++;
  }
}

void writePlotData() {
  chooseFilter();
  readDatasets();

  string full_filename_out = dir+filename_out;  
  FILE* fout = fopen(full_filename_out.c_str(), "w");
  assert(fout);

  writePlotData(fout);

  fclose(fout);
}

int main(int argc, char** argv) {
  if ( argc >= 8 ) {
    dir = string(argv[1]);
    filename_orig = string(argv[2]);
    filename_noisy = string(argv[3]);
    filename_out = string(argv[4]);
    noiseType = NoiseType(atoi(argv[5]));
    gridType = GridType(atoi(argv[6]));
    isWithBoundary = bool(atoi(argv[7]));
  }
  else {
    cout << "Some parameters are missing or incorrect. Please try again. " << endl;
    return -1;
  }

  writePlotData();

  return 0;
}
