/****************************************************************************

Author: Tai Meng

compiled and tested on amber.csil.sfu.ca

Usage: > ./plotNeighborhood [numNeighborhoods] [out.dat]
       - [numNeighborhoods]
         - specifies the number of neighborhoods to plot data points for
       - [out.dat]  
         - specifies the name of data file to be output
       - output in: 
         - ./[out.dat]
         - data points in ASCII for plotting in Matlab
         - data file format
           - line 1: numCols numRows (i.e. 2 50)
           - subsequent numRows lines: 1 line (or row) per entry
           - each entry has format
             - cc_radius cc_neighborHoodsize 
             - i.e. 1.000 7 
           - subsequent numRows lines: 1 line (or row) per entry
           - each entry has format
             - bcc_radius bcc_neighborhoodSize

****************************************************************************/

//============================================================
//Includes
//============================================================

#include <cmath>
#include <assert.h>
#include <cstdlib>

#include <iostream>
#include <string>
using namespace std;

#include <vuML.h>
#include <vuVector.h>
#include <Dataset.h>
#include <Utilities.h>
#include <MedianFilter.h>
#include <VVector.h>

//============================================================
//Global objects
//============================================================

MedianSphericalFilterCC filter_cc;
MedianSphericalFilterBCC filter_bcc;
SphericalFilter* active_filter;

string filename;
int numNeighborhoods;

//============================================================
//Functions
//============================================================

double getScale(GridType gridType) {
  if ( gridType == CC ) return 1.0;
  //scale for CC 64^3 vs BCC 45x45x90
  //theoretically this number is sqrt(2.0)
  //but in practice 1 unit of distance in CC needs to be 
  //equal to 1 unit of distance in BCC, so slightly
  //different from 2.0
  return 89.0/63.0;
}

void writeNeighborhoodData(FILE* fout, GridType gridType) {
  if ( gridType == CC ) active_filter = &filter_cc;
  else active_filter = &filter_bcc;

  int num_entries_so_far = 0;
  int radiusSquared = 0;
  int prevNeighborhoodSize = 0;
  int currentNeighborhoodSize;
  while ( num_entries_so_far < numNeighborhoods ) {
    double radius = sqrt(double(radiusSquared));
    active_filter->setRadius(radius);

    currentNeighborhoodSize = active_filter->getNeighborhoodSize();
    
    //avoid duplicate entries
    if ( currentNeighborhoodSize > prevNeighborhoodSize ) {
      assert(EQUALS(radius, active_filter->getRadius()));      
      double scale = getScale(gridType);

      fprintf(fout, "%lf\t%i\n", radius/scale, currentNeighborhoodSize);
      num_entries_so_far++;

      prevNeighborhoodSize = currentNeighborhoodSize;
    }
    
    radiusSquared++;
  }
}

void writeNeighborhoodData() {
  
  FILE* fout = fopen(filename.c_str(), "w");
  assert(fout);

  //first write header: numCols numRows
  //[radius_cc, cc_#neighbors]...
  //followed by [radius_bcc, bcc_#neighbors]
  fprintf(fout, "%i\t%i\n", 2, numNeighborhoods);
  writeNeighborhoodData(fout, CC);
  writeNeighborhoodData(fout, BCC);

  fclose(fout);
}


int main(int argc, char** argv) {
  if ( argc >= 3 ) {
    numNeighborhoods = atoi(argv[1]);
    filename = string(argv[2]);
  }
  else {
    cout << "Some parameters are missing or incorrect. Please try again. " << endl;
    return -1;
  }

  writeNeighborhoodData();

  return 0;
}
